/*LICENSE_BEGIN
 * ******************************************************************
 * Copyright 2006 by the Dainamite Robocup-Team at the
 * DAI-Labor, TU-Berlin (www.dai-labor.de)
 * 
 * Members of Dainamite are:
 * Holger Endert
 * Robert Wetzker
 * Thomas Karbe
 * Axel Heßler
 * Philippe Bütter
 * Felix Brossmann
 * 
 * Licensed under the GNU GPL (General Public License) Free Software 
 * License, Version 2. You may use this file only in compliance with 
 * the License. You may obtain a copy of the License at:
 * 
 * http://www.fsf.org/licensing/licenses/gpl.txt
 * ******************************************************************
LICENSE_END*/
package robocup.component;

/**
 * This class defines different constants together with functions to identify 
 * them. Most constants are using java 1.5 requiring enumerations.
 */
public interface ControllerNumbers {

	/**
	 * These constants identify the different lines of the soccer-field.
	 * They are used by the parser.
	 */
	public static final int LINE_CENTER = 2000;

	public static final int LINE_BOTTOM = 2001;

	public static final int LINE_TOP = 2002;

	public static final int LINE_LEFT = 2003;

	public static final int LINE_RIGHT = 2004;

	// **** REFEREE MESSAGES ******//

	public enum REFEREE_MESSAGE {
		FOUL_OWN, FOUL_OTHER, PENALTY_FOUL_OWN, PENALTY_FOUL_OTHER, FREE_KICK_FAULT_OWN, FREE_KICK_FAULT_OTHER, PENALTY_ON_FIELD_OWN, PENALTY_ON_FIELD_OTHER, BACKPASS_OWN, BACKPASS_OTHER, CATCH_FAULT_OWN, CATCH_FAULT_OTHER, HALF_TIME, TIME_UP, TIME_UP_WITHOUT_A_TEAM, TIME_EXTENDED, DROP_BALL, OFFSIDE_OWN, OFFSIDE_OTHER, GOALIE_CATCH_BALL_OWN, GOALIE_CATCH_BALL_OTHER, UNKNOWN
	};

	/**
	 * Playmodes can be simple as well as complex (meta-playmodes), which can be
	 * a collection of others. See for instance META_STANDARD_OWN, which gathers
	 * all playmodes for standard situations that should be executed by the own 
	 * team.
	 */
	public enum PLAY_MODE {
		ALWAYS(null), BEFORE_KICK_OFF(null), TIME_OVER(null), PLAY_ON(null), KICK_OFF_OWN(
				null), KICK_OFF_OTHER(null), KICK_IN_OWN(null), KICK_IN_OTHER(
				null), INDIRECT_FREE_KICK_OWN(null), FREE_KICK_OWN(null), FREE_KICK_OTHER(
				null), INDIRECT_FREE_KICK_OTHER(null), CORNER_KICK_OWN(null), CORNER_KICK_OTHER(
				null), GOAL_KICK_OWN(null), GOAL_KICK_OTHER(null), GOAL_OWN(
				null), GOAL_OTHER(null), PENALTY_SETUP_OWN(null), PENALTY_SETUP_OTHER(
				null), PENALTY_READY_OWN(null), PENALTY_READY_OTHER(null), PENALTY_TAKEN_OWN(
				null), PENALTY_TAKEN_OTHER(null), PENALTY_MISS_OWN(null), PENALTY_MISS_OTHER(
				null), PENALTY_SCORE_OWN(null), PENALTY_SCORE_OTHER(null), PENALTY_WINNER_OWN(
				null), PENALTY_WINNER_OTHER(null), TACKLE_DEAD(null), UNKNOWN(
				null),

		META_STANDARD_OWN(new PLAY_MODE[] { FREE_KICK_OWN,
				INDIRECT_FREE_KICK_OWN, CORNER_KICK_OWN, GOAL_KICK_OWN,
				KICK_IN_OWN, KICK_OFF_OWN }),

		META_PENALTY_OWN(new PLAY_MODE[] { PENALTY_MISS_OWN, PENALTY_READY_OWN,
				PENALTY_SCORE_OWN, PENALTY_SETUP_OWN, PENALTY_TAKEN_OWN, }),

		META_PENALTY_OTHER(new PLAY_MODE[] { PENALTY_MISS_OTHER,
				PENALTY_READY_OTHER, PENALTY_SCORE_OTHER, PENALTY_SETUP_OTHER,
				PENALTY_TAKEN_OTHER, }),

		META_STANDARD_OTHER(new PLAY_MODE[] { FREE_KICK_OTHER,
				INDIRECT_FREE_KICK_OTHER, CORNER_KICK_OTHER, GOAL_KICK_OTHER,
				KICK_IN_OTHER, KICK_OFF_OTHER });

		private PLAY_MODE[] subPlayModes = null;

		PLAY_MODE(PLAY_MODE[] subModes) {
			this.subPlayModes = subModes;
		}

		/**
		 * tells if the meta play mode contains the simple one.
		 * 
		 * used for meta playmodes which aggregate different play modes.
		 * 
		 * @param mode -
		 *            a simple play mode
		 * @return true if contained or equals
		 */
		public boolean contains(PLAY_MODE mode) {
			if (this == ALWAYS || this == mode)
				return true;

			if (this.subPlayModes == null)
				return false;

			for (PLAY_MODE p : this.subPlayModes) {
				if (p == mode) {
					return true;
				}
			}
			return false;
		}
	}

	
	/**
	 * There are two valid view-qualities HIGH and LOW. These affect the visual
	 * information receivement step by a factor. Each mode therefore consists
	 * of the string, which is accepted by the server to set the corresponding
	 * mode, and the factor for calculating the arrival of the next visual
	 * information (this depends as well on view-width - see server manual for
	 * details).
	 */
	public enum VIEW_QUALITY {
		HIGH("high", 1), LOW("low", 0.5);

		private double factor;

		private String string;

		private VIEW_QUALITY(String string, double factor) {
			this.factor = factor;
			this.string = string;
		}

		/**
		 * @return Returns the factor.
		 */
		public double getFactor() {

			return this.factor;
		}

		/**
		 * @return Returns the string.
		 */
		public String getString() {

			return this.string;
		}
	};

	
	/**
	 * There are three valid VIEW_ANGLEs: narrow (45), normal (default: 90)
	 * and wide (180). These are represented by the following enumeration,
	 * which contains a string, which is accepted by the server when switching
	 * to a certain mode (using the ChangeViewMode-Action), and the factor,
	 * which affects the arrival of the next visual information.
	 */
	public enum VIEW_ANGLE {
		NARROW("narrow", 0.5), NORMAL("normal", 1.0), WIDE("wide", 2.0);

		private double factor;

		private String string;

		private VIEW_ANGLE(String string, double factor) {
			this.factor = factor;
			this.string = string;
		}

		/**
		 * @return Returns the factor.
		 */
		public double getFactor() {

			return this.factor;
		}

		/**
		 * @return Returns the string.
		 */
		public String getString() {

			return this.string;
		}

	};

	/**
	 * Stamina-Levels are used to household the stamina of an agent, such that
	 * a certain level is not under-run. Currently, the following thresholds
	 * are defined by the levels, which causes the agent to move slower, if the
	 * stamina is near or below the given value:<br>
	 * <ul>
	 * <li>MAX: 3900</li>
	 * <li>HIGH: 3200</li>
	 * <li>MEDIUM: 2400</li>
	 * <li>MEDIUM_LOW: 2400</li>
	 * <li>LOW: 1250</li>
	 * <li>ZERO: 0</li>
	 * </ul>
	 * @see DAInamiteConstants
	 */
	public enum STAMINA_LEVEL {
		MAX, HIGH, MEDIUM, MEDIUM_LOW, LOW, ZERO
	};
	
	/**
	 * Identifiers for a team. Uncertain means, that it is not really known,
	 * which team is the correct one. An example is ball-possession. Due to
	 * noise an agent can't accurately predict, which agent is fastest to
	 * the ball, or if he thinks two agents of different teams are equal fast,
	 * no correct statement can be made.
	 */
	public enum TEAM {
		WE, UNCERTAIN, OPPONENT
	};

	/**
	 * The intercept modes are used to tell how a point (ball, player) can be
	 * reached fastest.
	 * 
	 * @see robocup.component.situations.Movements
	 * 
	 * DASH_FW = forward dashing DASH_BW = backward dashing TURNING_FW =
	 * 1x-turning + dashing fw TURNING_BW = 1x-turning + dashing bw TURNING2 =
	 * 2x turning TURNING3 = 3x truning NO_INTERCEPT = no interception possible
	 * 
	 */
	public enum INTERCEPTMODE {
		DASH_FW, DASH_BW, TURNING1_FW, TURNING1_BW, TURNING2, TURNING3, NO_INTERCEPT
	};
}
