/*LICENSE_BEGIN
 * ******************************************************************
 * Copyright 2006 by the Dainamite Robocup-Team at the
 * DAI-Labor, TU-Berlin (www.dai-labor.de)
 * 
 * Members of Dainamite are:
 * Holger Endert
 * Robert Wetzker
 * Thomas Karbe
 * Axel Heßler
 * Philippe Bütter
 * Felix Brossmann
 * 
 * Licensed under the GNU GPL (General Public License) Free Software 
 * License, Version 2. You may use this file only in compliance with 
 * the License. You may obtain a copy of the License at:
 * 
 * http://www.fsf.org/licensing/licenses/gpl.txt
 * ******************************************************************
LICENSE_END*/
package robocup.component.geometry;

import static java.lang.Math.PI;
import static java.lang.Math.atan;
import static java.lang.Math.tan;

/**
 * This class models a straight line in IR. A straight line <code>l</code> is
 * represented by an equation of the form: y = mx + n.
 */
public class StraightLine {
	/**
	 * The gradient.
	 */
	private double _m;

	/**
	 * The intersection with the y axis.
	 */
	private double _n;

	/**
	 * "epsilon"
	 */
	private final static double E = 0.000001;

	/**
	 * Constructs a new straight line.
	 * 
	 * @param m
	 *            The gradient.
	 * @param n
	 *            The interception with the y axis.
	 */
	public StraightLine(double m, double n) {
		this._m = m;
		this._n = n;
	}

	/**
	 * Constructs a new straight line through <code>point</code> with a slope
	 * <code>angle</code>.
	 * 
	 * @param point
	 * @param angle
	 */
	public StraightLine(Vektor point, double angle) {
		if (angle == 0)
			angle = E;
		this._m = tan(angle);
		this._n = point.y - this._m * point.x;
	}

	/**
	 * Constructs a new straight line with 2 vektors.
	 * 
	 * @param q
	 *            first vektor.
	 * @param p
	 *            second vektor.
	 */
	public StraightLine(Vektor q, Vektor p) {
		// avoid division by zero
		if (p.x == q.x)
			q.x += E;
		double m = (p.y - q.y) / (p.x - q.x);
		double n = q.y - m * q.x;

		this._m = m;
		this._n = n;
	}

	/**
	 * This method calculates this line's distance to a given point.
	 * 
	 * @param point
	 *            The point.
	 * @return The distance.
	 */
	public double getDistanceToPoint(Vektor point) {
		return getDistanceToPoint(point.x, point.y);
	}

	/**
	 * This method calculates this line's distance to a given point.
	 * 
	 * @param point(x,y)
	 *            The point.
	 * @return The distance.
	 */
	public double getDistanceToPoint(double x, double y) {
		return getVektorToLine(new Vektor(Vektor.XY, x, y)).getLength();
	}

	/**
	 * This method calculates the shortest Vektor from a given point to a line.
	 * 
	 * @param point
	 * @return the perpendicular
	 */
	public Vektor getVektorToLine(Vektor point) {
		// special case: m=0
		if (this._m == 0)
			return new Vektor(Vektor.XY, 0, this._n - point.y);

		// find the perpendicular line to this line through point
		double n = point.x / this._m + point.y;
		StraightLine lot = new StraightLine(-1 / this._m, n);
		// find the intersection point of the two lines
		Vektor intersectionPoint = getIntersectionPoint(lot);
		Vektor vLot = intersectionPoint.sub(point);
		return vLot;
	}

	/**
	 * This method checks if the given Vektor is on line
	 * 
	 * @param point
	 * @return
	 */
	public boolean isOnLine(Vektor point) {
		if ((point.y <= this._m * point.x + this._n + E)
				&& (point.y >= this._m * point.x + this._n - E))
			return true;
		return false;
	}

	/**
	 * This method calculates the intersection of this line and another given
	 * line.
	 * 
	 * @param line
	 *            the given line.
	 * @return The intersection point, or null if both lines are parallel.
	 */
	public Vektor getIntersectionPoint(StraightLine line) {
		double m = line.getM();
		// if the lines are parallel, i don't feel like searching for the
		// intersection point...
		if (this._m == m)
			return null;
		// avoid division by zero
		if (m == 0)
			m = E;
		double mNew = this._m;
		double nNew = this._n;
		double intersectionPointX, intersectionPointY;
		nNew -= line.getN();
		mNew /= m;
		nNew /= m;
		intersectionPointX = nNew / (1 - mNew);
		intersectionPointY = this._m * intersectionPointX + this._n;
		return new Vektor(Vektor.XY, intersectionPointX, intersectionPointY);
	}

	/**
	 * @return Returns m.
	 */
	public double getM() {
		return this._m;
	}

	/**
	 * @return Returns n.
	 */
	public double getN() {
		return this._n;
	}

	/**
	 * The slope of a line in .
	 * 
	 * @return
	 */
	public double getSlope() {
		return atan(this._m) * (180.0 / PI);
	}

	/**
	 * Returns a string representation of this line.
	 */
	@Override
	public String toString() {
		String line = "-";
		if (this._m < 0)
			line = "\\";
		if (this._m > 0)
			line = "/";
		String sign = this._n < 0 ? "" : "+ ";
		return "y = " + this._m + " * x " + sign + this._n + " " + line;
	}
}
