/*LICENSE_BEGIN
 * ******************************************************************
 * All original parts of SoccerScope2003 have copyright (2003) by
 * Team YowAI, represented by Mr. Shuhei Shiota
 * The Department of Creative Informatics
 * The Graduate School of Information Science and Technology
 * The University of Tokyo
 * 
 * Additional classes (except libraries in /lib) were developed by 
 * the Dainamite Robocup-Team of the DAI-Labor, TU-Berlin 
 * (www.dai-labor.de)
 * 
 * Extended/new classes/packages are:
 *   soccerscope.analyzer.LogfileAnalyzer
 *   soccerscope.daistatistic.*
 *   soccerscope.daistatistic.charts.*
 *   soccerscope.daistatistic.tools.*
 *   soccerscope.db.*
 *   soccerscope.dialog.*
 *   soccerscope.model.PlayerWorldModel
 *   soccerscope.model.SimplePlayer
 *   soccerscope.net.AgentConnector
 *   soccerscope.training.*
 *   soccerscope.training.conditions.*
 * ******************************************************************
LICENSE_END*/
package soccerscope.file;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipInputStream;

import soccerscope.model.Ball;
import soccerscope.model.ClientScene;
import soccerscope.model.Param;
import soccerscope.model.PlayMode;
import soccerscope.model.SceneBuilder;
import soccerscope.model.Team;

public class LogFileReader extends SceneBuilder {

	public final static int BUFFER_SIZE = 2048;

	private InputStream in;

	private InputStreamReader isr;

	private BufferedReader br;

	private int version;

	private String filename;

	private final static int REC_VERSION_1 = 1;

	private final static int REC_VERSION_2 = 2;

	private final static int REC_VERSION_3 = 3;

	public LogFileReader(String name) throws IOException {

		FileInputStream fin = new FileInputStream(name);
		if (name.endsWith(".zip")) {
			ZipInputStream zin = new ZipInputStream(fin);
			// ZipEntry zen = zin.getNextEntry();
			this.in = new BufferedInputStream(zin);
		} else if (name.endsWith(".gz")) {
			this.in = new BufferedInputStream(new GZIPInputStream(fin));
		} else {
			this.in = new BufferedInputStream(fin);
		}

		this.filename = name;
		if (name.endsWith(".rcl") || name.endsWith(".rcl.gz")) {
			this.isr = new InputStreamReader(this.in);
			this.br = new BufferedReader(this.isr);
			this.type = LOGFILE | LOGFILE_VER_3;
		} else {
			// read logfile header
			byte[] header = new byte[4];
			try {
				this.in.read(header);
			} catch (IOException ie) {
				throw ie;
			}

			if (header[0] == 'U' && header[1] == 'L' && header[2] == 'G') {
				this.version = header[3];
			} else {
				throw new IOException("no version info: " + name);
			}

			switch (this.version) {
			case REC_VERSION_1:
				this.type = LOGFILE | LOGFILE_VER_1;
				break;
			case REC_VERSION_2:
				this.type = LOGFILE | LOGFILE_VER_2;
				break;
			case REC_VERSION_3:
				this.type = LOGFILE | LOGFILE_VER_3;
				break;
			default:
				throw new IOException("invalid version: " + this.version + " "
						+ name);
			}
			System.err.print("logfile version" + this.version + ": " + name);
		}
	}

	public String getFilename() {
		return this.filename;
	}

	public int getVersion() {
		return this.version;
	}

	public void close() {
		try {
			this.in.close();
			System.err.println("...done");
		} catch (IOException ie) {
		}
	}

	public int readLog(byte[] buffer) throws IOException {
		try {
			return readPacket(buffer);
		} catch (IOException ie) {
			throw ie;
		}
	}

	@Override
	public int readPacket(byte[] buffer) throws IOException {
		switch (this.version) {
		case REC_VERSION_1:
			return -1;

		case REC_VERSION_2:
			try {
				return readLogVer2(buffer);
			} catch (IOException ie) {
				throw ie;
			}

		case REC_VERSION_3:
			try {
				return readLogVer3(buffer);
			} catch (IOException ie) {
				throw ie;
			}

		default:
			return -1;
		}
	}

	// Reading Server Log
	public String readServerLog(/* String line */) throws IOException {
		return this.br.readLine();
	}

	public void closeServerLog() {
		try {
			this.br.close();
		} catch (IOException ie) {
		}
	}

	// logging protocol ver3
	private int readLogVer3(byte[] buffer) throws IOException {
		// read mode
		try {
			if (this.in.read(buffer, 0, 2) == -1) {
				return -1;
			}
		} catch (IOException ie) {
			throw ie;
		}

		int mode = readShort(buffer, 0);

		if (mode == SHOW_MODE) {
			try {
				if (this.in.read(buffer, 2, 1428) == -1) {
					return -1;
				}
				return 2 + 1428;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == MSG_MODE) {
			try {
				// read board info & length of message
				if (this.in.read(buffer, 2, 4) == -1) {
					return -1;
				}

				int len = buffer[4];
				len = readShort(buffer, 4);

				if (this.in.read(buffer, 6, len) == -1) {
					return -1;
				}
				return 2 + 2 + 2 + len;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == PM_MODE) {
			try {
				if (this.in.read(buffer, 2, 1) == -1) {
					return -1;
				}
				return 2 + 1;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == TEAM_MODE) {
			try {
				if (this.in.read(buffer, 2, 36) == -1) {
					return -1;
				}
				return 2 + 36;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == PT_MODE) {
			try {
				if (this.in.read(buffer, 2, 88) == -1) {
					return -1;
				}
				return 2 + 88;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == PARAM_MODE) {
			try {
				if (this.in.read(buffer, 2, 396) == -1) {
					return -1;
				}
				return 2 + 396;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == PPARAM_MODE) {
			try {
				if (this.in.read(buffer, 2, 132) == -1) {
					return -1;
				}
				return 2 + 132;
			} catch (IOException ie) {
				throw ie;
			}
		}

		return 0;
	}

	// logging protocol ver2
	private int readLogVer2(byte[] buffer) throws IOException {
		// read mode
		try {
			if (this.in.read(buffer, 0, 2) == -1) {
				return -1;
			}
		} catch (IOException ie) {
			throw ie;
		}

		int mode = readShort(buffer, 0);

		if (mode == SHOW_MODE) {
			try {
				if (this.in.read(buffer, 2, 316) == -1) {
					return -1;
				}
				return 2 + 316;
			} catch (IOException ie) {
				throw ie;
			}
		} else if (mode == MSG_MODE) {
			try {
				// read board info & length of message
				if (this.in.read(buffer, 2, 4) == -1) {
					return -1;
				}

				int len = buffer[4];
				len = readShort(buffer, 4);

				if (this.in.read(buffer, 6, len) == -1) {
					return -1;
				}
				return 2 + 2 + 2 + len;
			} catch (IOException ie) {
				throw ie;
			}
		}
		return 0;
	}

	@Override
	public ClientScene makeScene(byte[] packet, PlayMode playmode, Team left,
			Team right) {
		ClientScene clientScene = new ClientScene();
		int offset = 0;

		if (this.version == REC_VERSION_3) {
			clientScene.left = new Team(left);
			clientScene.right = new Team(right);
			clientScene.pmode = new PlayMode(playmode);

			offset = 2;
			// read ball_t
			clientScene.ball.mode = Ball.STAND;
			// // x, y (long)
			clientScene.ball.pos.x = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;
			clientScene.ball.pos.y = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;
			// // deltax, deltay (long)
			clientScene.ball.vel.x = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;
			clientScene.ball.vel.y = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;

			// read player_t
			for (int i = 0; i < Param.MAX_PLAYER * 2; i++) {
				// // mode (short)
				clientScene.player[i].mode = readShort(packet, offset);
				offset += 2;

				// // unum
				if (i < Param.MAX_PLAYER) {
					clientScene.player[i].unum = i + 1;
					clientScene.player[i].side = Team.LEFT_SIDE;
				} else {
					clientScene.player[i].unum = i - Param.MAX_PLAYER + 1;
					clientScene.player[i].side = Team.RIGHT_SIDE;
				}

				// // type (short)
				clientScene.player[i].type = readShort(packet, offset);
				offset += 2;

				// // x, y (long)
				clientScene.player[i].pos.x = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;
				clientScene.player[i].pos.y = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;

				// // deltax, deltay (long)
				clientScene.player[i].vel.x = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;
				clientScene.player[i].vel.y = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;

				// // body_angle (long) (radians)
				clientScene.player[i].angle = (float) Math.toDegrees(readLong(
						packet, offset)
						/ SHOWINFO_SCALE2);
				offset += 4;

				// // head_angle (long) (radians)
				clientScene.player[i].angleNeck = (int) Math
						.toDegrees(readLong(packet, offset) / SHOWINFO_SCALE2);
				offset += 4;

				// // view_width (long) (radians)
				clientScene.player[i].angleVisible = (int) Math
						.toDegrees(readLong(packet, offset) / SHOWINFO_SCALE2);
				offset += 4;

				// // view_quality (short)
				clientScene.player[i].viewQuality = readShort(packet, offset);
				offset += 4;

				// // stamina (long)
				clientScene.player[i].stamina = (int) (readLong(packet, offset) / SHOWINFO_SCALE2);
				offset += 4;

				// // effort (long)
				// clientScene.player[i].effort = readLong(packet, offset) /
				// SHOWINFO_SCALE2;
				offset += 4;

				// // recovery (long)
				// clientScene.player[i].recovery = readLong(packet, offset) /
				// SHOWINFO_SCALE2;
				offset += 4;

				// // kick_count (short)
				clientScene.player[i].kickCount = readShort(packet, offset);
				offset += 2;

				// // dash_count (short)
				clientScene.player[i].dashCount = readShort(packet, offset);
				offset += 2;

				// // turn_count (short)
				clientScene.player[i].turnCount = readShort(packet, offset);
				offset += 2;

				// // say_count (short)
				clientScene.player[i].sayCount = readShort(packet, offset);
				offset += 2;

				// // turnneck_count (short)
				clientScene.player[i].turnNeckCount = readShort(packet, offset);
				offset += 2;

				// // catch_count (short)
				clientScene.player[i].catchCount = readShort(packet, offset);
				offset += 2;

				// // move_count (short)
				clientScene.player[i].moveCount = readShort(packet, offset);
				offset += 2;

				// // chg_view_count (short)
				clientScene.player[i].changeViewCount = readShort(packet,
						offset);
				offset += 2;
			}
		} else if (this.version == REC_VERSION_2) {
			offset = 2;
			PlayMode pplaymode = makePlayMode(packet);
			offset += 2;

			Team tleft = makeLeftTeam(packet, offset);
			Team tright = makeRightTeam(packet, offset);
			offset += 36;

			clientScene.left = new Team(tleft);
			clientScene.right = new Team(tright);
			clientScene.pmode = new PlayMode(pplaymode);

			// read ball info (pos_t[0])
			// // enable (short)
			clientScene.ball.mode = Ball.STAND;
			offset += 2;

			// // side (short)
			offset += 2;

			// // unum (short)
			offset += 2;

			// // angle (short)
			offset += 2;

			// // x, y (short)
			clientScene.ball.pos.x = readShort(packet, offset) / SHOWINFO_SCALE;
			offset += 2;
			clientScene.ball.pos.y = readShort(packet, offset) / SHOWINFO_SCALE;
			offset += 2;

			// read pos_t[1-22]
			for (int i = 0; i < Param.MAX_PLAYER * 2; i++) {
				int enable;
				int side;
				int unum;
				int index;

				// // enable (short)
				enable = readShort(packet, offset);
				offset += 2;

				// // side (short)
				side = readShort(packet, offset);
				offset += 2;

				// // unum (short)
				unum = readShort(packet, offset);
				offset += 2;

				if (side == Team.LEFT_SIDE)
					index = (unum - 1);
				else
					index = (unum - 1) + Param.MAX_PLAYER;

				clientScene.player[index].unum = unum;
				clientScene.player[index].side = side;
				clientScene.player[index].mode = enable;

				// // angle (short)
				clientScene.player[index].angle = readShort(packet, offset);
				offset += 2;

				// // x, y (short)
				clientScene.player[index].pos.x = readShort(packet, offset)
						/ SHOWINFO_SCALE;
				offset += 2;
				clientScene.player[index].pos.y = readShort(packet, offset)
						/ SHOWINFO_SCALE;
				offset += 2;
			}
		}

		// read time (short)
		clientScene.time = readShort(packet, offset);
		offset += 2;

		// check offside
		clientScene.left.offsideline = 0;
		clientScene.right.offsideline = 0;
		float leftx[] = new float[Param.MAX_PLAYER + 1];
		float rightx[] = new float[Param.MAX_PLAYER + 1];
		for (int i = 0; i < Param.MAX_PLAYER; i++) {
			leftx[i] = clientScene.player[i].pos.x;
			rightx[i] = clientScene.player[i + Param.MAX_PLAYER].pos.x;
		}
		leftx[Param.MAX_PLAYER] = clientScene.ball.pos.x;
		rightx[Param.MAX_PLAYER] = clientScene.ball.pos.x;
		Arrays.sort(leftx);
		Arrays.sort(rightx);
		if (leftx[1] < 0)
			clientScene.left.offsideline = leftx[1];
		if (rightx[Param.MAX_PLAYER - 1] > 0)
			clientScene.right.offsideline = rightx[Param.MAX_PLAYER - 1];

		clientScene.left.offside = false;
		clientScene.right.offside = false;
		for (int i = 0; i < Param.MAX_PLAYER; i++) {
			if (clientScene.player[i].pos.x > clientScene.right.offsideline) {
				clientScene.player[i].offside = true;
				clientScene.left.offside = true;
			} else {
				clientScene.player[i].offside = false;
			}
			if (clientScene.player[i + Param.MAX_PLAYER].pos.x < clientScene.left.offsideline) {
				clientScene.player[i + Param.MAX_PLAYER].offside = true;
				clientScene.right.offside = true;
			} else {
				clientScene.player[i + Param.MAX_PLAYER].offside = false;
			}
		}

		return clientScene;
	}

}
