/*LICENSE_BEGIN
 * ******************************************************************
 * All original parts of SoccerScope2003 have copyright (2003) by
 * Team YowAI, represented by Mr. Shuhei Shiota
 * The Department of Creative Informatics
 * The Graduate School of Information Science and Technology
 * The University of Tokyo
 * 
 * Additional classes (except libraries in /lib) were developed by 
 * the Dainamite Robocup-Team of the DAI-Labor, TU-Berlin 
 * (www.dai-labor.de)
 * 
 * Extended/new classes/packages are:
 *   soccerscope.analyzer.LogfileAnalyzer
 *   soccerscope.daistatistic.*
 *   soccerscope.daistatistic.charts.*
 *   soccerscope.daistatistic.tools.*
 *   soccerscope.db.*
 *   soccerscope.dialog.*
 *   soccerscope.model.PlayerWorldModel
 *   soccerscope.model.SimplePlayer
 *   soccerscope.net.AgentConnector
 *   soccerscope.training.*
 *   soccerscope.training.conditions.*
 * ******************************************************************
LICENSE_END*/
package soccerscope.net;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.Arrays;

import soccerscope.model.Ball;
import soccerscope.model.ClientScene;
import soccerscope.model.Param;
import soccerscope.model.PlayMode;
import soccerscope.model.SceneBuilder;
import soccerscope.model.Team;

public class SoccerServerConnection extends SceneBuilder {

	public static int DEFAULT_PORT = 6000;

	public static String DEFAULT_HOST = "localhost";

	public static int DEFAULT_PROTOCOL = SceneBuilder.MONITOR_PROTOCOL_2;

	private InetAddress address;

	private String host;

	private int port;

	private DatagramSocket socket;

	public SoccerServerConnection() throws UnknownHostException,
			SocketException {
		this(DEFAULT_HOST, DEFAULT_PORT, SERVER | DEFAULT_PROTOCOL);
	}

	public SoccerServerConnection(String hostname) throws UnknownHostException,
			SocketException {
		this(hostname, DEFAULT_PORT, SERVER | DEFAULT_PROTOCOL);
	}

	public SoccerServerConnection(String hostname, int port, int protocol)
			throws UnknownHostException, SocketException {
		this.type = SERVER | protocol;
		this.host = hostname;
		this.port = port;
		try {
			this.address = InetAddress.getByName(hostname);
		} catch (UnknownHostException uhe) {
			throw uhe;
		}
		try {
			this.socket = new DatagramSocket();
		} catch (SocketException se) {
			throw se;
		}
	}

	public String getHostName() {
		return this.host;
	}

	public int getPort() {
		return this.port;
	}

	private void sendMessage(String msg) throws IOException {
		byte[] rawMsg = msg.getBytes();
		DatagramPacket packet = new DatagramPacket(rawMsg, rawMsg.length,
				this.address, this.port);
		try {
			this.socket.send(packet);
		} catch (IOException ie) {
			throw ie;
		}
	}

	public void dispinit() throws IOException {
		String msg = new String("(dispinit)");
		if ((this.type & MONITOR_PROTOCOL_2) == MONITOR_PROTOCOL_2) {
			msg = new String("(dispinit version 2)");
		}
		try {
			sendMessage(msg);
		} catch (IOException ie) {
			throw ie;
		}
	}

	public void dispbye() throws IOException {
		try {
			sendMessage("(dispbye)");
		} catch (IOException ie) {
			throw ie;
		}
	}

	public void dispstart() throws IOException {
		try {
			sendMessage("(dispstart)");
		} catch (IOException ie) {
			throw ie;
		}
	}

	public void dispfoul(float x, float y, int side) throws IOException {
		try {
			sendMessage("(dispfoul " + (int) (x * SHOWINFO_SCALE) + " "
					+ (int) (y * SHOWINFO_SCALE) + " " + side + ")");
		} catch (IOException ie) {
			throw ie;
		}
	}

	public void dispdiscard(int side, int unum) throws IOException {
		try {
			sendMessage("(dispdiscard " + side + " " + unum + ")");
		} catch (IOException ie) {
			throw ie;
		}
	}

	public void dispplayer(int side, int unum, float x, float y, int angle)
			throws IOException {
		try {
			sendMessage("(dispplayer " + side + " " + unum + " "
					+ (int) (x * SHOWINFO_SCALE) + " "
					+ (int) (y * SHOWINFO_SCALE) + " " + angle + ")");
		} catch (IOException ie) {
			throw ie;
		}
	}

	@Override
	public int readPacket(byte[] packet) throws IOException {
		DatagramPacket udppacket = new DatagramPacket(packet, packet.length);
		try {
			this.socket.receive(udppacket);
		} catch (IOException ie) {
			throw ie;
		}
		return udppacket.getLength();
	}

	@Override
	public ClientScene makeScene(byte[] packet, PlayMode playmode, Team left,
			Team right) {
		ClientScene clientScene = new ClientScene();
		int offset = 0;

		if ((this.type & MONITOR_PROTOCOL_2) == MONITOR_PROTOCOL_2) {
			offset = 4;
			clientScene.pmode = new PlayMode();
			clientScene.pmode.pmode = readChar(packet, offset);
			offset += 2;
			clientScene.left = makeLeftTeam(packet, offset);
			clientScene.right = makeRightTeam(packet, offset);
			offset += 36;

			offset += 2;

			// read ball_t
			clientScene.ball.mode = Ball.STAND;
			// // x, y (long)
			clientScene.ball.pos.x = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;
			clientScene.ball.pos.y = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;
			// // deltax, deltay (long)
			clientScene.ball.vel.x = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;
			clientScene.ball.vel.y = readLong(packet, offset) / SHOWINFO_SCALE2;
			offset += 4;

			// read player_t
			for (int i = 0; i < Param.MAX_PLAYER * 2; i++) {
				// // mode (short)
				clientScene.player[i].mode = readShort(packet, offset);
				offset += 2;

				// // unum
				if (i < Param.MAX_PLAYER) {
					clientScene.player[i].unum = i + 1;
					clientScene.player[i].side = Team.LEFT_SIDE;
				} else {
					clientScene.player[i].unum = i - Param.MAX_PLAYER + 1;
					clientScene.player[i].side = Team.RIGHT_SIDE;
				}

				// // type (short)
				clientScene.player[i].type = readShort(packet, offset);
				offset += 2;

				// // x, y (long)
				clientScene.player[i].pos.x = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;
				clientScene.player[i].pos.y = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;

				// // deltax, deltay (long)
				clientScene.player[i].vel.x = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;
				clientScene.player[i].vel.y = readLong(packet, offset)
						/ SHOWINFO_SCALE2;
				offset += 4;

				// // body_angle (long) (radians)
				clientScene.player[i].angle = (int) Math.toDegrees(readLong(
						packet, offset)
						/ SHOWINFO_SCALE2);
				offset += 4;

				// // head_angle (long) (radians)
				clientScene.player[i].angleNeck = (int) Math
						.toDegrees(readLong(packet, offset) / SHOWINFO_SCALE2);
				offset += 4;

				// // view_width (long) (radians)
				clientScene.player[i].angleVisible = (int) Math
						.toDegrees(readLong(packet, offset) / SHOWINFO_SCALE2);
				offset += 4;

				// // view_quality (short)
				clientScene.player[i].viewQuality = readShort(packet, offset);
				offset += 4;

				// // stamina (long)
				clientScene.player[i].stamina = (int) (readLong(packet, offset) / SHOWINFO_SCALE2);
				offset += 4;

				// // effort (long)
				// clientScene.player[i].effort = readLong(packet, offset) /
				// Param.SHOWINFO_SCALE2;
				offset += 4;

				// // recovery (long)
				// clientScene.player[i].recovery = readLong(packet, offset) /
				// Param.SHOWINFO_SCALE2;
				offset += 4;

				// // kick_count (short)
				clientScene.player[i].kickCount = readShort(packet, offset);
				offset += 2;

				// // dash_count (short)
				clientScene.player[i].dashCount = readShort(packet, offset);
				offset += 2;

				// // turn_count (short)
				clientScene.player[i].turnCount = readShort(packet, offset);
				offset += 2;

				// // say_count (short)
				clientScene.player[i].sayCount = readShort(packet, offset);
				offset += 2;

				// // turnneck_count (short)
				clientScene.player[i].turnNeckCount = readShort(packet, offset);
				offset += 2;

				// // catch_count (short)
				clientScene.player[i].catchCount = readShort(packet, offset);
				offset += 2;

				// // move_count (short)
				clientScene.player[i].moveCount = readShort(packet, offset);
				offset += 2;

				// // chg_view_count (short)
				clientScene.player[i].changeViewCount = readShort(packet,
						offset);
				offset += 2;
			}

		} else if ((this.type & MONITOR_PROTOCOL_1) == MONITOR_PROTOCOL_1) {
			offset = 2;
			PlayMode pplaymode = makePlayMode(packet);
			offset += 2;

			Team tleft = makeLeftTeam(packet, offset);
			Team tright = makeRightTeam(packet, offset);
			offset += 36;

			clientScene.left = new Team(tleft);
			clientScene.right = new Team(tright);
			clientScene.pmode = new PlayMode(pplaymode);

			// read ball info (pos_t[0])
			// // enable (short)
			clientScene.ball.mode = Ball.STAND;
			offset += 2;

			// // side (short)
			offset += 2;

			// // unum (short)
			offset += 2;

			// // angle (short)
			offset += 2;

			// // x, y (short)
			clientScene.ball.pos.x = readShort(packet, offset) / SHOWINFO_SCALE;
			offset += 2;
			clientScene.ball.pos.y = readShort(packet, offset) / SHOWINFO_SCALE;
			offset += 2;

			// read pos_t[1-22]
			for (int i = 0; i < Param.MAX_PLAYER * 2; i++) {
				int enable;
				int side;
				int unum;
				int index;

				// // enable (short)
				enable = readShort(packet, offset);
				offset += 2;

				// // side (short)
				side = readShort(packet, offset);
				offset += 2;

				// // unum (short)
				unum = readShort(packet, offset);
				offset += 2;

				if (side == Team.LEFT_SIDE)
					index = (unum - 1);
				else
					index = (unum - 1) + Param.MAX_PLAYER;

				clientScene.player[index].unum = unum;
				clientScene.player[index].side = side;
				clientScene.player[index].mode = enable;

				// // angle (short)
				clientScene.player[index].angle = readShort(packet, offset);
				offset += 2;

				// // x, y (short)
				clientScene.player[index].pos.x = readShort(packet, offset)
						/ SHOWINFO_SCALE;
				offset += 2;
				clientScene.player[index].pos.y = readShort(packet, offset)
						/ SHOWINFO_SCALE;
				offset += 2;
			}
		}
		// read time (short)
		clientScene.time = readUnsignedShort(packet, offset);
		offset += 2;

		// check offside
		clientScene.left.offsideline = 0;
		clientScene.right.offsideline = 0;
		float leftx[] = new float[Param.MAX_PLAYER + 1];
		float rightx[] = new float[Param.MAX_PLAYER + 1];
		for (int i = 0; i < Param.MAX_PLAYER; i++) {
			leftx[i] = clientScene.player[i].pos.x;
			rightx[i] = clientScene.player[i + Param.MAX_PLAYER].pos.x;
		}
		leftx[Param.MAX_PLAYER] = clientScene.ball.pos.x;
		rightx[Param.MAX_PLAYER] = clientScene.ball.pos.x;
		Arrays.sort(leftx);
		Arrays.sort(rightx);
		if (leftx[1] < 0)
			clientScene.left.offsideline = leftx[1];
		if (rightx[Param.MAX_PLAYER - 1] > 0)
			clientScene.right.offsideline = rightx[Param.MAX_PLAYER - 1];

		clientScene.left.offside = false;
		clientScene.right.offside = false;
		for (int i = 0; i < Param.MAX_PLAYER; i++) {
			if (clientScene.player[i].pos.x > clientScene.right.offsideline) {
				clientScene.player[i].offside = true;
				clientScene.left.offside = true;
			} else {
				clientScene.player[i].offside = false;
			}
			if (clientScene.player[i + Param.MAX_PLAYER].pos.x < clientScene.left.offsideline) {
				clientScene.player[i + Param.MAX_PLAYER].offside = true;
				clientScene.right.offside = true;
			} else {
				clientScene.player[i + Param.MAX_PLAYER].offside = false;
			}
		}

		return clientScene;
	}

}
